import os, subprocess, gzip, bz2, shutil, re, signal
from ConfigParser import SafeConfigParser

from scm import *

class git(scm):

    name = 'git'

    features = { SCM_INIT : True,
                 SCM_CLONE : True,
                 SCM_PATHS : True,
                 SCM_PUSH : True,
                 SCM_PULL : True,
                 SCM_INCOMING : True,
                 SCM_OUTGOING : True,
                 SCM_CREATE_BUNDLE : True,
                 SCM_IMPORT_BUNDLE : True,
                 SCM_EXPORT_PATCH : False,
                 SCM_IMPORT_PATCH : False,
                 SCM_SERVER : True,
                 SCM_ARCHIVE : True,
                 SCM_EXECUTE : True,
                 SCM_TAG : True,
                 SCM_UPDATE : True,
                 SCM_ANNOTATE : True,
                 SCM_OPTIMIZE : True,
                 SCM_ALT_TYPES : False, }

    archive_formats = { ARCHIVE_TAR : True,
                        ARCHIVE_TAR_BZ2 : True,
                        ARCHIVE_TAR_GZ : True,
                        ARCHIVE_ZIP_UNCOMPRESSED : True,
                        ARCHIVE_ZIP_COMPRESSED : True }
    
    update_modes = { UPDATE_TO_TIP : True,
                     UPDATE_TO_REV : True,
                     UPDATE_TO_TIME : False, }
    
    def __init__(self):
        scm.__init__(self)

        self.serverpid = 0

    @staticmethod
    def probe():
        if os.path.exists('.git'):
            print 'Detected a git repository'
            return git()
        else:
            return None

    def execute(self, cmd, args):
        arg_str = ''
        for a in args:
            arg_str += ' ' + a

        output = os.popen('git ' + cmd + str(arg_str)).read()
        return output

    def optimize(self):
        self.execute('gc', [])

    def annotate(self, file):
        annotated_lines = []

        # Need to remove path of repository of the start of file, or
        # git will (bizarrely) give us the commit details of the
        # repository's first commit, regardless of when the file came
        # from.  :-S
        cwd = os.getcwd()
        if cwd != '/':
            cwd = cwd + '/'
        file = file[len(cwd):]

        text = self.execute('annotate', ['HEAD', file])

        annotate_re = re.compile(r"(?P<revid>\S*)\s*\((?P<user>.*)\s*(?P<date>\d\d\d\d-\d\d-\d\d "
                                 + r"\d\d:\d\d:\d\d .\d\d\d\d)\s*\d*\)(?P<text>.*$)")

        for line in text.split('\n'):
            m = annotate_re.match(line)
            if m == None:
                print 'Failed to match!'
                print 'line was: %s' % line
                continue
            user = m.group('user')
            revid = m.group('revid')
            date = m.group('date')
            text = m.group('text')

            ann_line = annotatedLine(user, revid, None, date, None, text)
            annotated_lines.append(ann_line)

        return annotated_lines

    def update_to_tip(self):
        self.execute('checkout', ['master'])

    def update_to_rev(self, rev):
        self.execute('checkout', [rev])

    def init(self, path):
        if os.path.exists(path) == False:
            os.makedirs(path)
        os.chdir(path)
        self.execute('init', [])        

    def clone(self, source_path, new_path):
        self.execute('clone', [source_path, new_path])

    def paths(self):
        path_list = []
        
        config = open('.git/config') # failure case!  other config files influence?

        branch_name = None

        # This could maybe be a little more paranoid about oddly
        # formatted config files but it should do the job for the
        # moment.

        for line in config:
            branch_match = re.match(r"\[remote\s*\"(.*)\"\]", line)
            if branch_match != None:
                branch_name = branch_match.group(1)
                print "Set branch name %s" % branch_name
                continue

            if re.match(r"\[.*\]", line):
                branch_name = None
                continue

            url_match = re.match(r"\s*url\s*=\s*(\S*)\s*", line)
            if url_match != None and branch_name != None:
                path_list.append((branch_name, url_match.group(1)))
                continue

        return path_list

    def push(self, path):
        print 'git push'
        self.execute('push', [path])

    def pull(self, path):
        print 'git pull'
        self.execute('pull', [path])

    def branches(self):
        branches = []
        branch_output = self.execute('branch', ['-a'])
        branch_lines = branch_output.split('\n')
        for l in branch_lines:
            m = re.match(r"[\* ] (.*)$", l)
            if m == None:
                break
            branches.append(m.group(1))
        return branches

    def incoming(self, path):
        # In the unlikely event the user may have created this branch,
        # leave it alone.
        for b in self.branches():
            if b == 'pmpu-incoming':
                return 'ERROR: pmpu-incoming branch already exists.'

        print 'Fetching incoming changesets from master branch of %s' % path
        self.execute('fetch', [path, 'master:pmpu-incoming'])
        print 'Generating incoming list...'
        in_csets = self.execute('log', ['master..pmpu-incoming'])
        print 'Removing temporary pmpu-incoming branch'
        self.execute('branch', ['-D', 'pmpu-incoming'])
        return in_csets

    def outgoing(self, path):
        # In the unlikely event the user may have created this branch,
        # leave it alone.
        for b in self.branches():
            if b == 'pmpu-outgoing':
                return 'ERROR: pmpu-outgoing branch already exists.'

        print 'Fetching master branch of %s' % path
        self.execute('fetch', [path, 'master:pmpu-outgoing'])
        print 'Generating outgoing list...'
        out_csets = self.execute('log', ['pmpu-outgoing..master'])
        self.execute('branch', ['-D', 'pmpu-outgoing'])
        return out_csets

    def create_bundle(self, bundle_path, destPath):
        # In the unlikely event the user may have created this branch,
        # leave it alone.
        for b in self.branches():
            if b == 'pmpu-outgoing':
                print 'Failed to create bundle: pmpu-outgoing already exists'
                return # Return failure notification!
        print 'Fetching master branch of %s' % destPath
        self.execute('fetch', [destPath, 'master:pmpu-outgoing'])
        print 'Creating bundle file %s' % bundle_path
        self.execute('bundle', ['create', bundle_path, 'pmpu-outgoing..master'])
        print 'Removing temporary pmpu-outgoing branch'
        self.execute('branch', ['-D', 'pmpu-outgoing'])

    def import_bundle(self, bundle_path):
        print 'Importing bundle'
        self.execute('pull', [bundle_path, 'master'])

    def export_patch(self):
        raise scmException('Export patch not supported')

    def import_patch(self):
        raise scmException('Import patch not supported')

    def archive(self, path, type):
        path = str(path) # force QString to python string

        cmdline = ['git', 'archive']

        format_arg = ''
        if type == ARCHIVE_TAR:
            cmdline.append('--format=tar')
            outfile = open(path, 'w')
        elif type == ARCHIVE_TAR_GZ:
            cmdline.append('--format=tar')
            outfile = gzip.GzipFile(path, 'w')
        elif type == ARCHIVE_TAR_BZ2:
            cmdline.append('--format=tar')
            outfile = bz2.BZ2File(path, 'w')
        elif type == ARCHIVE_ZIP_COMPRESSED:
            cmdline.append('--format=zip')
            cmdline.append('-9')
            outfile = open(path, 'w')
        elif type == ARCHIVE_ZIP_UNCOMPRESSED:
            cmdline.append('--format=zip')
            cmdline.append('-0')
            outfile = open(path, 'w')

        cmdline.append('HEAD')

        p1 = subprocess.Popen(cmdline, stdout=subprocess.PIPE)

        shutil.copyfileobj(p1.stdout, outfile)
        p1.wait()

        outfile.close()

    def start_server(self):
        self.serverpid = os.fork()
        if not self.serverpid:
            os.execlp('git', 'git', 'daemon', '--export-all', '--base-path=.')

    def stop_server(self):
        if self.serverpid:
            os.kill(self.serverpid, signal.SIGTERM)
            self.serverpid = 0
            os.wait()

    def tag(self, name):
        self.execute('tag', [name])

    def destroy(self):
        self.stop_server()


register_backend(git)

